# -*- coding: utf-8 -*-
"""
docstring goes here.

:copyright: Copyright 2014 by the Elephant team, see AUTHORS.txt.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

from itertools import chain
import unittest

from neo.test.generate_datasets import fake_neo, get_fake_values
from neo.test.tools import assert_same_sub_schema
from numpy.testing.utils import assert_array_equal

import elephant.neo_tools as nt


# A list of neo object attributes that contain arrays.
ARRAY_ATTRS = ['waveforms',
               'times',
               'durations',
               'labels',
               'channel_indexes',
               'channel_names'
               ]


def strip_iter_values(targ, array_attrs=ARRAY_ATTRS):
    """Remove iterable, non-string values from a dictionary.

    `elephant.neo_tools.extract_neo_attrs` automatically strips out
    non-scalar values from attributes.  This function does the same to a
    manually-extracted dictionary.

    Parameters
    ----------
    targ : dict
           The dictionary of values to process.
    array_attrs : list of str objects, optional
                  The list of attribute names to remove.  If not specified,
                  uses `elephant.test.test_neo_tools.ARRAY_ATTRS`.

    Returns
    -------
    dict
        A copy of `targ` with the target values (if present) removed.

    Notes
    -----

    Always returns a copy, even if nothing was removed.

    This function has the values to remove hard-coded.  This is intentional
    to make sure that `extract_neo_attrs` is removing all the attributes
    it is supposed to and only the attributes it is supposed to.  Please do
    NOT change this to any sort of automatic detection, if it is missing
    values please add them manually.

    """
    targ = targ.copy()

    for attr in array_attrs:
        targ.pop(attr, None)
    return targ


class GetAllObjsTestCase(unittest.TestCase):
    def test__get_all_objs__float_valueerror(self):
        value = 5.
        with self.assertRaises(ValueError):
            nt._get_all_objs(value, 'Block')

    def test__get_all_objs__list_float_valueerror(self):
        value = [5.]
        with self.assertRaises(ValueError):
            nt._get_all_objs(value, 'Block')

    def test__get_all_objs__epoch_for_event_valueerror(self):
        value = fake_neo('Epoch', n=10, seed=0)
        with self.assertRaises(ValueError):
            nt._get_all_objs(value, 'Event')

    def test__get_all_objs__empty_list(self):
        targ = []
        value = []

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_nested_list(self):
        targ = []
        value = [[], [[], [[]]]]

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_dict(self):
        targ = []
        value = {}

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_nested_dict(self):
        targ = []
        value = {'a': {}, 'b': {'c': {}, 'd': {'e': {}}}}

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_itert(self):
        targ = []
        value = iter([])

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_nested_iter(self):
        targ = []
        value = iter([iter([]), iter([iter([]), iter([iter([])])])])

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__empty_nested_many(self):
        targ = []
        value = iter([[], {'c': [], 'd':(iter([]),)}])

        res = nt._get_all_objs(value, 'Block')

        self.assertEqual(targ, res)

    def test__get_all_objs__spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0)]
        value = fake_neo('SpikeTrain', n=10, seed=0)

        res = nt._get_all_objs(value, 'SpikeTrain')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__list_spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0),
                fake_neo('SpikeTrain', n=10, seed=1)]
        value = [fake_neo('SpikeTrain', n=10, seed=0),
                 fake_neo('SpikeTrain', n=10, seed=1)]

        res = nt._get_all_objs(value, 'SpikeTrain')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__nested_list_epoch(self):
        targ = [fake_neo('Epoch', n=10, seed=0),
                fake_neo('Epoch', n=10, seed=1)]
        value = [[fake_neo('Epoch', n=10, seed=0)],
                 fake_neo('Epoch', n=10, seed=1)]

        res = nt._get_all_objs(value, 'Epoch')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__iter_spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0),
                fake_neo('SpikeTrain', n=10, seed=1)]
        value = iter([fake_neo('SpikeTrain', n=10, seed=0),
                      fake_neo('SpikeTrain', n=10, seed=1)])

        res = nt._get_all_objs(value, 'SpikeTrain')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__nested_iter_epoch(self):
        targ = [fake_neo('Epoch', n=10, seed=0),
                fake_neo('Epoch', n=10, seed=1)]
        value = iter([iter([fake_neo('Epoch', n=10, seed=0)]),
                      fake_neo('Epoch', n=10, seed=1)])

        res = nt._get_all_objs(value, 'Epoch')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__dict_spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0),
                fake_neo('SpikeTrain', n=10, seed=1)]
        value = {'a': fake_neo('SpikeTrain', n=10, seed=0),
                 'b': fake_neo('SpikeTrain', n=10, seed=1)}

        res = nt._get_all_objs(value, 'SpikeTrain')

        self.assertEqual(len(targ), len(res))
        for i, itarg in enumerate(targ):
            for ires in res:
                if itarg.annotations['seed'] == ires.annotations['seed']:
                    assert_same_sub_schema(itarg, ires)
                    break
            else:
                raise ValueError('Target %s not in result' % i)

    def test__get_all_objs__nested_dict_spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0),
                fake_neo('SpikeTrain', n=10, seed=1)]
        value = {'a': fake_neo('SpikeTrain', n=10, seed=0),
                 'b': {'c': fake_neo('SpikeTrain', n=10, seed=1)}}

        res = nt._get_all_objs(value, 'SpikeTrain')

        self.assertEqual(len(targ), len(res))
        for i, itarg in enumerate(targ):
            for ires in res:
                if itarg.annotations['seed'] == ires.annotations['seed']:
                    assert_same_sub_schema(itarg, ires)
                    break
            else:
                raise ValueError('Target %s not in result' % i)

    def test__get_all_objs__nested_many_spiketrain(self):
        targ = [fake_neo('SpikeTrain', n=10, seed=0),
                fake_neo('SpikeTrain', n=10, seed=1)]
        value = {'a': [fake_neo('SpikeTrain', n=10, seed=0)],
                 'b': iter([fake_neo('SpikeTrain', n=10, seed=1)])}

        res = nt._get_all_objs(value, 'SpikeTrain')

        self.assertEqual(len(targ), len(res))
        for i, itarg in enumerate(targ):
            for ires in res:
                if itarg.annotations['seed'] == ires.annotations['seed']:
                    assert_same_sub_schema(itarg, ires)
                    break
            else:
                raise ValueError('Target %s not in result' % i)

    def test__get_all_objs__unit_spiketrain(self):
        value = fake_neo('Unit', n=3, seed=0)
        targ = [fake_neo('SpikeTrain', n=3, seed=train.annotations['seed'])
                for train in value.spiketrains]

        for train in value.spiketrains:
            train.annotations.pop('i', None)
            train.annotations.pop('j', None)

        res = nt._get_all_objs(value, 'SpikeTrain')

        assert_same_sub_schema(targ, res)

    def test__get_all_objs__block_epoch(self):
        value = fake_neo('Block', n=3, seed=0)
        targ = [fake_neo('Epoch', n=3, seed=train.annotations['seed'])
                for train in value.list_children_by_class('Epoch')]

        for epoch in value.list_children_by_class('Epoch'):
            epoch.annotations.pop('i', None)
            epoch.annotations.pop('j', None)

        res = nt._get_all_objs(value, 'Epoch')

        assert_same_sub_schema(targ, res)


class ExtractNeoAttrsTestCase(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.block = fake_neo('Block', seed=0)

    def assert_dicts_equal(self, d1, d2):
        """Assert that two dictionaries are equal, taking into account arrays.

        Normally, `unittest.TestCase.assertEqual` doesn't work with
        dictionaries containing arrays.  This works around that.

        Parameters
        ----------

        d1, d2 : dict
                 The dictionaries to compare

        Returns
        -------

        Nothing

        Raises
        ------
        AssertionError : If the `d1` and `d2` are not equal.

        """
        try:
            self.assertEqual(d1, d2)
        except ValueError:
            for (key1, value1), (key2, value2) in zip(sorted(d1.items()),
                                                      sorted(d2.items())):
                self.assertEqual(key1, key2)
                try:
                    if hasattr(value1, 'keys') and hasattr(value2, 'keys'):
                        self.assert_dicts_equal(value1, value2)
                    elif hasattr(value1, 'dtype') and hasattr(value2, 'dtype'):
                        assert_array_equal(value1, value2)
                    else:
                        self.assertEqual(value1, value2)
                except BaseException as exc:
                    exc.args += ('key: %s' % key1,)
                    raise

    def test__extract_neo_attrs__spiketrain_noarray(self):
        obj = fake_neo('SpikeTrain', seed=0)
        targ = get_fake_values('SpikeTrain', seed=0)
        targ = strip_iter_values(targ)

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res11 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)

        self.assertEqual(targ, res00)
        self.assertEqual(targ, res10)
        self.assertEqual(targ, res20)
        self.assertEqual(targ, res01)
        self.assertEqual(targ, res11)
        self.assertEqual(targ, res21)

    def test__extract_neo_attrs__spiketrain_noarray_skip_none(self):
        obj = fake_neo('SpikeTrain', seed=0)
        targ = get_fake_values('SpikeTrain', seed=0)
        targ = strip_iter_values(targ)
        for key, value in targ.copy().items():
            if value is None:
                del targ[key]

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     skip_none=True)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True, skip_none=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False, skip_none=True)
        res01 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                     skip_none=True)
        res11 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True, skip_none=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False, skip_none=True)

        self.assertEqual(targ, res00)
        self.assertEqual(targ, res10)
        self.assertEqual(targ, res20)
        self.assertEqual(targ, res01)
        self.assertEqual(targ, res11)
        self.assertEqual(targ, res21)

    def test__extract_neo_attrs__epoch_noarray(self):
        obj = fake_neo('Epoch', seed=0)
        targ = get_fake_values('Epoch', seed=0)
        targ = strip_iter_values(targ)

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res11 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)

        self.assertEqual(targ, res00)
        self.assertEqual(targ, res10)
        self.assertEqual(targ, res20)
        self.assertEqual(targ, res01)
        self.assertEqual(targ, res11)
        self.assertEqual(targ, res21)

    def test__extract_neo_attrs__event_noarray(self):
        obj = fake_neo('Event', seed=0)
        targ = get_fake_values('Event', seed=0)
        targ = strip_iter_values(targ)

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res11 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                     child_first=False)

        self.assertEqual(targ, res00)
        self.assertEqual(targ, res10)
        self.assertEqual(targ, res20)
        self.assertEqual(targ, res01)
        self.assertEqual(targ, res11)
        self.assertEqual(targ, res21)

    def test__extract_neo_attrs__spiketrain_parents_empty_array(self):
        obj = fake_neo('SpikeTrain', seed=0)
        targ = get_fake_values('SpikeTrain', seed=0)
        del targ['times']

        res000 = nt.extract_neo_attrs(obj, parents=False)
        res100 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res200 = nt.extract_neo_attrs(obj, parents=False, child_first=False)
        res010 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res110 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=True)
        res210 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=False)
        res001 = nt.extract_neo_attrs(obj, parents=True)
        res101 = nt.extract_neo_attrs(obj, parents=True, child_first=True)
        res201 = nt.extract_neo_attrs(obj, parents=True, child_first=False)
        res011 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res111 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=True)
        res211 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=False)

        self.assert_dicts_equal(targ, res000)
        self.assert_dicts_equal(targ, res100)
        self.assert_dicts_equal(targ, res200)
        self.assert_dicts_equal(targ, res010)
        self.assert_dicts_equal(targ, res110)
        self.assert_dicts_equal(targ, res210)
        self.assert_dicts_equal(targ, res001)
        self.assert_dicts_equal(targ, res101)
        self.assert_dicts_equal(targ, res201)
        self.assert_dicts_equal(targ, res011)
        self.assert_dicts_equal(targ, res111)
        self.assert_dicts_equal(targ, res211)

    def test__extract_neo_attrs__epoch_parents_empty_array(self):
        obj = fake_neo('Epoch', seed=0)
        targ = get_fake_values('Epoch', seed=0)

        res000 = nt.extract_neo_attrs(obj, parents=False)
        res100 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res200 = nt.extract_neo_attrs(obj, parents=False, child_first=False)
        res010 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res110 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=True)
        res210 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=False)
        res001 = nt.extract_neo_attrs(obj, parents=True)
        res101 = nt.extract_neo_attrs(obj, parents=True, child_first=True)
        res201 = nt.extract_neo_attrs(obj, parents=True, child_first=False)
        res011 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res111 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=True)
        res211 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=False)

        self.assert_dicts_equal(targ, res000)
        self.assert_dicts_equal(targ, res100)
        self.assert_dicts_equal(targ, res200)
        self.assert_dicts_equal(targ, res010)
        self.assert_dicts_equal(targ, res110)
        self.assert_dicts_equal(targ, res210)
        self.assert_dicts_equal(targ, res001)
        self.assert_dicts_equal(targ, res101)
        self.assert_dicts_equal(targ, res201)
        self.assert_dicts_equal(targ, res011)
        self.assert_dicts_equal(targ, res111)
        self.assert_dicts_equal(targ, res211)

    def test__extract_neo_attrs__event_parents_empty_array(self):
        obj = fake_neo('Event', seed=0)
        targ = get_fake_values('Event', seed=0)

        res000 = nt.extract_neo_attrs(obj, parents=False)
        res100 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res200 = nt.extract_neo_attrs(obj, parents=False, child_first=False)
        res010 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res110 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=True)
        res210 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                      child_first=False)
        res001 = nt.extract_neo_attrs(obj, parents=True)
        res101 = nt.extract_neo_attrs(obj, parents=True, child_first=True)
        res201 = nt.extract_neo_attrs(obj, parents=True, child_first=False)
        res011 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res111 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=True)
        res211 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                      child_first=False)

        self.assert_dicts_equal(targ, res000)
        self.assert_dicts_equal(targ, res100)
        self.assert_dicts_equal(targ, res200)
        self.assert_dicts_equal(targ, res010)
        self.assert_dicts_equal(targ, res110)
        self.assert_dicts_equal(targ, res210)
        self.assert_dicts_equal(targ, res001)
        self.assert_dicts_equal(targ, res101)
        self.assert_dicts_equal(targ, res201)
        self.assert_dicts_equal(targ, res011)
        self.assert_dicts_equal(targ, res111)
        self.assert_dicts_equal(targ, res211)

    def test__extract_neo_attrs__spiketrain_noparents_noarray(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        targ = get_fake_values('SpikeTrain', seed=obj.annotations['seed'])
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=True)
        res2 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res1['i']
        del res2['i']
        del res0['j']
        del res1['j']
        del res2['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)
        self.assertEqual(targ, res2)

    def test__extract_neo_attrs__epoch_noparents_noarray(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        targ = get_fake_values('Epoch', seed=obj.annotations['seed'])
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=True)
        res2 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res1['i']
        del res2['i']
        del res0['j']
        del res1['j']
        del res2['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)
        self.assertEqual(targ, res2)

    def test__extract_neo_attrs__event_noparents_noarray(self):
        obj = self.block.list_children_by_class('Event')[0]
        targ = get_fake_values('Event', seed=obj.annotations['seed'])
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=False, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=True)
        res2 = nt.extract_neo_attrs(obj, parents=False, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res1['i']
        del res2['i']
        del res0['j']
        del res1['j']
        del res2['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)
        self.assertEqual(targ, res2)

    def test__extract_neo_attrs__spiketrain_noparents_array(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        targ = get_fake_values('SpikeTrain', seed=obj.annotations['seed'])
        del targ['times']

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=False)
        res11 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, child_first=False)

        del res00['i']
        del res10['i']
        del res20['i']
        del res01['i']
        del res11['i']
        del res21['i']
        del res00['j']
        del res10['j']
        del res20['j']
        del res01['j']
        del res11['j']
        del res21['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res20)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)
        self.assert_dicts_equal(targ, res21)

    def test__extract_neo_attrs__epoch_noparents_array(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        targ = get_fake_values('Epoch', seed=obj.annotations['seed'])

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=False)
        res11 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, child_first=False)

        del res00['i']
        del res10['i']
        del res20['i']
        del res01['i']
        del res11['i']
        del res21['i']
        del res00['j']
        del res10['j']
        del res20['j']
        del res01['j']
        del res11['j']
        del res21['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res20)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)
        self.assert_dicts_equal(targ, res21)

    def test__extract_neo_attrs__event_noparents_array(self):
        obj = self.block.list_children_by_class('Event')[0]
        targ = get_fake_values('Event', seed=obj.annotations['seed'])

        res00 = nt.extract_neo_attrs(obj, parents=False, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=True)
        res20 = nt.extract_neo_attrs(obj, parents=False, skip_array=False,
                                     child_first=False)
        res01 = nt.extract_neo_attrs(obj, parents=False)
        res11 = nt.extract_neo_attrs(obj, parents=False, child_first=True)
        res21 = nt.extract_neo_attrs(obj, parents=False, child_first=False)

        del res00['i']
        del res10['i']
        del res20['i']
        del res01['i']
        del res11['i']
        del res21['i']
        del res00['j']
        del res10['j']
        del res20['j']
        del res01['j']
        del res11['j']
        del res21['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res20)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)
        self.assert_dicts_equal(targ, res21)

    def test__extract_neo_attrs__spiketrain_parents_childfirst_noarray(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        blk = self.block
        seg = self.block.segments[0]
        rcg = self.block.recordingchannelgroups[0]
        unit = self.block.recordingchannelgroups[0].units[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('RecordingChannelGroup',
                                    seed=rcg.annotations['seed']))
        targ.update(get_fake_values('Unit', seed=unit.annotations['seed']))
        targ.update(get_fake_values('SpikeTrain',
                                    seed=obj.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=True)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)

    def test__extract_neo_attrs__epoch_parents_childfirst_noarray(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Epoch', seed=obj.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=True)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)

    def test__extract_neo_attrs__event_parents_childfirst_noarray(self):
        obj = self.block.list_children_by_class('Event')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Event', seed=obj.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True)
        res1 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=True)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']

        self.assertEqual(targ, res0)
        self.assertEqual(targ, res1)

    def test__extract_neo_attrs__spiketrain_parents_parentfirst_noarray(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        blk = self.block
        seg = self.block.segments[0]
        rcg = self.block.recordingchannelgroups[0]
        unit = self.block.recordingchannelgroups[0].units[0]

        targ = get_fake_values('SpikeTrain', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Unit', seed=unit.annotations['seed']))
        targ.update(get_fake_values('RecordingChannelGroup',
                                    seed=rcg.annotations['seed']))
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res0['j']

        self.assertEqual(targ, res0)

    def test__extract_neo_attrs__epoch_parents_parentfirst_noarray(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Epoch', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res0['j']

        self.assertEqual(targ, res0)

    def test__extract_neo_attrs__event_parents_parentfirst_noarray(self):
        obj = self.block.list_children_by_class('Event')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Event', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))
        targ = strip_iter_values(targ)

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=True,
                                    child_first=False)

        del res0['i']
        del res0['j']

        self.assertEqual(targ, res0)

    def test__extract_neo_attrs__spiketrain_parents_childfirst_array(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        blk = self.block
        seg = self.block.segments[0]
        rcg = self.block.recordingchannelgroups[0]
        unit = self.block.recordingchannelgroups[0].units[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('RecordingChannelGroup',
                                    seed=rcg.annotations['seed']))
        targ['channel_names'] = rcg.channel_names
        targ.update(get_fake_values('Unit', seed=unit.annotations['seed']))
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('SpikeTrain',
                                    seed=obj.annotations['seed']))
        del targ['times']

        res00 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                     child_first=True)
        res01 = nt.extract_neo_attrs(obj, parents=True)
        res11 = nt.extract_neo_attrs(obj, parents=True,  child_first=True)

        del res00['i']
        del res10['i']
        del res01['i']
        del res11['i']
        del res00['j']
        del res10['j']
        del res01['j']
        del res11['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)

    def test__extract_neo_attrs__epoch_parents_childfirst_array(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Epoch', seed=obj.annotations['seed']))

        res00 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                     child_first=True)
        res01 = nt.extract_neo_attrs(obj, parents=True)
        res11 = nt.extract_neo_attrs(obj, parents=True,  child_first=True)

        del res00['i']
        del res10['i']
        del res01['i']
        del res11['i']
        del res00['j']
        del res10['j']
        del res01['j']
        del res11['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)

    def test__extract_neo_attrs__event_parents_childfirst_array(self):
        obj = self.block.list_children_by_class('Event')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Block', seed=blk.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Event', seed=obj.annotations['seed']))

        res00 = nt.extract_neo_attrs(obj, parents=True, skip_array=False)
        res10 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                     child_first=True)
        res01 = nt.extract_neo_attrs(obj, parents=True)
        res11 = nt.extract_neo_attrs(obj, parents=True,  child_first=True)

        del res00['i']
        del res10['i']
        del res01['i']
        del res11['i']
        del res00['j']
        del res10['j']
        del res01['j']
        del res11['j']

        self.assert_dicts_equal(targ, res00)
        self.assert_dicts_equal(targ, res10)
        self.assert_dicts_equal(targ, res01)
        self.assert_dicts_equal(targ, res11)

    def test__extract_neo_attrs__spiketrain_parents_parentfirst_array(self):
        obj = self.block.list_children_by_class('SpikeTrain')[0]
        blk = self.block
        seg = self.block.segments[0]
        rcg = self.block.recordingchannelgroups[0]
        unit = self.block.recordingchannelgroups[0].units[0]

        targ = get_fake_values('SpikeTrain', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Unit', seed=unit.annotations['seed']))
        targ.update(get_fake_values('RecordingChannelGroup',
                                    seed=rcg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))
        del targ['times']
        del targ['channel_indexes']
        del targ['channel_names']

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                    child_first=False)
        res1 = nt.extract_neo_attrs(obj, parents=True, child_first=False)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']
        del res0['channel_indexes']
        del res1['channel_indexes']
        del res0['channel_names']
        del res1['channel_names']

        self.assert_dicts_equal(targ, res0)
        self.assert_dicts_equal(targ, res1)

    def test__extract_neo_attrs__epoch_parents_parentfirst_array(self):
        obj = self.block.list_children_by_class('Epoch')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Epoch', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                    child_first=False)
        res1 = nt.extract_neo_attrs(obj, parents=True, child_first=False)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']

        self.assert_dicts_equal(targ, res0)
        self.assert_dicts_equal(targ, res1)

    def test__extract_neo_attrs__event_parents_parentfirst_array(self):
        obj = self.block.list_children_by_class('Event')[0]
        blk = self.block
        seg = self.block.segments[0]

        targ = get_fake_values('Event', seed=obj.annotations['seed'])
        targ.update(get_fake_values('Segment', seed=seg.annotations['seed']))
        targ.update(get_fake_values('Block', seed=blk.annotations['seed']))

        res0 = nt.extract_neo_attrs(obj, parents=True, skip_array=False,
                                    child_first=False)
        res1 = nt.extract_neo_attrs(obj, parents=True, child_first=False)

        del res0['i']
        del res1['i']
        del res0['j']
        del res1['j']

        self.assert_dicts_equal(targ, res0)
        self.assert_dicts_equal(targ, res1)


class GetAllSpiketrainsTestCase(unittest.TestCase):
    def test__get_all_spiketrains__spiketrain(self):
        obj = fake_neo('SpikeTrain', seed=0, n=5)
        res0 = nt.get_all_spiketrains(obj)

        targ = obj

        self.assertEqual(1, len(res0))

        assert_same_sub_schema(targ, res0[0])

    def test__get_all_spiketrains__unit(self):
        obj = fake_neo('Unit', seed=0, n=7)
        obj.spiketrains.append(obj.spiketrains[0])
        res0 = nt.get_all_spiketrains(obj)

        targ = fake_neo('Unit', seed=0, n=7).spiketrains

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__segment(self):
        obj = fake_neo('Segment', seed=0, n=5)
        obj.spiketrains.extend(obj.spiketrains)
        res0 = nt.get_all_spiketrains(obj)

        targ = fake_neo('Segment', seed=0, n=5).spiketrains

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__block(self):
        obj = fake_neo('Block', seed=0, n=3)
        iobj1 = obj.recordingchannelgroups[0].units[0]
        obj.recordingchannelgroups[0].units.append(iobj1)
        iobj2 = obj.recordingchannelgroups[0].units[2].spiketrains[1]
        obj.recordingchannelgroups[1].units[1].spiketrains.append(iobj2)
        res0 = nt.get_all_spiketrains(obj)

        targ = fake_neo('Block', seed=0, n=3)
        targ = targ.list_children_by_class('SpikeTrain')

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__list(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].recordingchannelgroups[0].units[0]
        obj[2].recordingchannelgroups[0].units.append(iobj1)
        iobj2 = obj[1].recordingchannelgroups[1].units[2].spiketrains[1]
        obj[2].recordingchannelgroups[0].units[1].spiketrains.append(iobj2)
        obj.append(obj[-1])
        res0 = nt.get_all_spiketrains(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('SpikeTrain') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__tuple(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].recordingchannelgroups[0].units[0]
        obj[2].recordingchannelgroups[0].units.append(iobj1)
        iobj2 = obj[1].recordingchannelgroups[1].units[2].spiketrains[1]
        obj[2].recordingchannelgroups[0].units[1].spiketrains.append(iobj2)
        obj.append(obj[0])
        res0 = nt.get_all_spiketrains(tuple(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('SpikeTrain') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__iter(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].recordingchannelgroups[0].units[0]
        obj[2].recordingchannelgroups[0].units.append(iobj1)
        iobj2 = obj[1].recordingchannelgroups[1].units[2].spiketrains[1]
        obj[2].recordingchannelgroups[0].units[1].spiketrains.append(iobj2)
        obj.append(obj[1])
        res0 = nt.get_all_spiketrains(iter(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('SpikeTrain') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_spiketrains__dict(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].recordingchannelgroups[0].units[0]
        obj[2].recordingchannelgroups[0].units.append(iobj1)
        iobj2 = obj[1].recordingchannelgroups[1].units[2].spiketrains[1]
        obj[2].recordingchannelgroups[0].units[1].spiketrains.append(iobj2)
        obj.append(obj[1])
        obj = dict((i, iobj) for i, iobj in enumerate(obj))
        res0 = nt.get_all_spiketrains(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('SpikeTrain') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)


class GetAllEventsTestCase(unittest.TestCase):
    def test__get_all_events__event(self):
        obj = fake_neo('Event', seed=0, n=5)
        res0 = nt.get_all_events(obj)

        targ = obj

        self.assertEqual(1, len(res0))

        assert_same_sub_schema(targ, res0[0])

    def test__get_all_events__segment(self):
        obj = fake_neo('Segment', seed=0, n=5)
        obj.events.extend(obj.events)
        res0 = nt.get_all_events(obj)

        targ = fake_neo('Segment', seed=0, n=5).events

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_events__block(self):
        obj = fake_neo('Block', seed=0, n=3)
        iobj1 = obj.segments[0]
        obj.segments.append(iobj1)
        iobj2 = obj.segments[0].events[1]
        obj.segments[1].events.append(iobj2)
        res0 = nt.get_all_events(obj)

        targ = fake_neo('Block', seed=0, n=3)
        targ = targ.list_children_by_class('Event')

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_events__list(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].events[1]
        obj[2].segments[1].events.append(iobj2)
        obj.append(obj[-1])
        res0 = nt.get_all_events(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Event') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_events__tuple(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].events[1]
        obj[2].segments[1].events.append(iobj2)
        obj.append(obj[0])
        res0 = nt.get_all_events(tuple(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Event') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_events__iter(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].events[1]
        obj[2].segments[1].events.append(iobj2)
        obj.append(obj[0])
        res0 = nt.get_all_events(iter(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Event') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_events__dict(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].events[1]
        obj[2].segments[1].events.append(iobj2)
        obj.append(obj[0])
        obj = dict((i, iobj) for i, iobj in enumerate(obj))
        res0 = nt.get_all_events(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Event') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)


class GetAllEpochsTestCase(unittest.TestCase):
    def test__get_all_epochs__epoch(self):
        obj = fake_neo('Epoch', seed=0, n=5)
        res0 = nt.get_all_epochs(obj)

        targ = obj

        self.assertEqual(1, len(res0))

        assert_same_sub_schema(targ, res0[0])

    def test__get_all_epochs__segment(self):
        obj = fake_neo('Segment', seed=0, n=5)
        obj.epochs.extend(obj.epochs)
        res0 = nt.get_all_epochs(obj)

        targ = fake_neo('Segment', seed=0, n=5).epochs

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_epochs__block(self):
        obj = fake_neo('Block', seed=0, n=3)
        iobj1 = obj.segments[0]
        obj.segments.append(iobj1)
        iobj2 = obj.segments[0].epochs[1]
        obj.segments[1].epochs.append(iobj2)
        res0 = nt.get_all_epochs(obj)

        targ = fake_neo('Block', seed=0, n=3)
        targ = targ.list_children_by_class('Epoch')

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_epochs__list(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].epochs[1]
        obj[2].segments[1].epochs.append(iobj2)
        obj.append(obj[-1])
        res0 = nt.get_all_epochs(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Epoch') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_epochs__tuple(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].epochs[1]
        obj[2].segments[1].epochs.append(iobj2)
        obj.append(obj[0])
        res0 = nt.get_all_epochs(tuple(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Epoch') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_epochs__iter(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].epochs[1]
        obj[2].segments[1].epochs.append(iobj2)
        obj.append(obj[0])
        res0 = nt.get_all_epochs(iter(obj))

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Epoch') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)

    def test__get_all_epochs__dict(self):
        obj = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        obj.append(obj[-1])
        iobj1 = obj[2].segments[0]
        obj[2].segments.append(iobj1)
        iobj2 = obj[1].segments[2].epochs[1]
        obj[2].segments[1].epochs.append(iobj2)
        obj.append(obj[0])
        obj = dict((i, iobj) for i, iobj in enumerate(obj))
        res0 = nt.get_all_epochs(obj)

        targ = [fake_neo('Block', seed=i, n=3) for i in range(3)]
        targ = [iobj.list_children_by_class('Epoch') for iobj in targ]
        targ = list(chain.from_iterable(targ))

        self.assertTrue(len(res0) > 0)

        self.assertEqual(len(targ), len(res0))

        assert_same_sub_schema(targ, res0)


if __name__ == '__main__':
    unittest.main()



'''
Basic processing procedures for analog signals (e.g., performing a z-score of a signal, or filtering a signal).

:copyright: Copyright 2014-2015 by the Elephant team, see AUTHORS.txt.
:license: Modified BSD, see LICENSE.txt for details.
'''

from __future__ import division, print_function
import numpy as np
import scipy.signal
import quantities as pq
import neo


def zscore(signal, inplace=True):
    '''
    Apply a z-score operation to one or several AnalogSignalArray objects.

    The z-score operation subtracts the mean :math:`\\mu` of the signal, and
    divides by its standard deviation :math:`\\sigma`:

    .. math::
         Z(x(t))= \\frac{x(t)-\\mu}{\\sigma}

    If an AnalogSignalArray containing multiple signals is provided, the
    z-transform is always calculated for each signal individually.

    If a list of AnalogSignalArray objects is supplied, the mean and standard
    deviation are calculated across all objects of the list. Thus, all list
    elements are z-transformed by the same values of :math:`\\mu` and
    :math:`\\sigma`. For AnalogSignalArrays, each signal of the array is
    treated separately across list elements. Therefore, the number of signals
    must be identical for each AnalogSignalArray of the list.

    Parameters
    ----------
    signal : neo.AnalogSignalArray or list of neo.AnalogSignalArray
        Signals for which to calculate the z-score.
    inplace : bool
        If True, the contents of the input signal(s) is replaced by the
        z-transformed signal. Otherwise, a copy of the original
        AnalogSignalArray(s) is returned. Default: True

    Returns
    -------
    neo.AnalogSignalArray or list of neo.AnalogSignalArray
        The output format matches the input format: for each supplied
        AnalogSignalArray object a corresponding object is returned containing
        the z-transformed signal with the unit dimensionless.

    Use Case
    --------
    You may supply a list of AnalogSignalArray objects, where each object in
    the list contains the data of one trial of the experiment, and each signal
    of the AnalogSignalArray corresponds to the recordings from one specific
    electrode in a particular trial. In this scenario, you will z-transform the
    signal of each electrode separately, but transform all trials of a given
    electrode in the same way.

    Examples
    --------
    >>> a = neo.AnalogSignalArray(
    ...       np.array([1, 2, 3, 4, 5, 6]).reshape(-1,1)*mV,
    ...       t_start=0*s, sampling_rate=1000*Hz)

    >>> b = neo.AnalogSignalArray(
    ...       np.transpose([[1, 2, 3, 4, 5, 6], [11, 12, 13, 14, 15, 16]])*mV,
    ...       t_start=0*s, sampling_rate=1000*Hz)

    >>> c = neo.AnalogSignalArray(
    ...       np.transpose([[21, 22, 23, 24, 25, 26], [31, 32, 33, 34, 35, 36]])*mV,
    ...       t_start=0*s, sampling_rate=1000*Hz)

    >>> print zscore(a)
    [[-1.46385011]
     [-0.87831007]
     [-0.29277002]
     [ 0.29277002]
     [ 0.87831007]
     [ 1.46385011]] dimensionless

    >>> print zscore(b)
    [[-1.46385011 -1.46385011]
     [-0.87831007 -0.87831007]
     [-0.29277002 -0.29277002]
     [ 0.29277002  0.29277002]
     [ 0.87831007  0.87831007]
     [ 1.46385011  1.46385011]] dimensionless

    >>> print zscore([b,c])   #  doctest: +NORMALIZE_WHITESPACE
    [<AnalogSignalArray(array([[-1.11669108, -1.08361877],
       [-1.0672076 , -1.04878252],
       [-1.01772411, -1.01394628],
       [-0.96824063, -0.97911003],
       [-0.91875714, -0.94427378],
       [-0.86927366, -0.90943753]]) * dimensionless, [0.0 s, 0.006 s],
       sampling rate: 1000.0 Hz)>,
       <AnalogSignalArray(array([[ 0.78170952,  0.84779261],
       [ 0.86621866,  0.90728682],
       [ 0.9507278 ,  0.96678104],
       [ 1.03523694,  1.02627526],
       [ 1.11974608,  1.08576948],
       [ 1.20425521,  1.1452637 ]]) * dimensionless, [0.0 s, 0.006 s],
       sampling rate: 1000.0 Hz)>]
    '''
    # Transform input to a list
    if type(signal) is not list:
        signal = [signal]

    # Calculate mean and standard deviation
    m = np.mean(np.concatenate(signal), axis=0, keepdims=True)
    s = np.std(np.concatenate(signal), axis=0, keepdims=True)

    if not inplace:
        # Create new signal instance
        result = [sig.duplicate_with_new_array(
            (sig.magnitude - m.magnitude) / s.magnitude) for sig in signal]
        for sig in result:
            sig /= sig.units
    else:
        # Overwrite signal
        for sig in signal:
            sig[:] = pq.Quantity(
                (sig.magnitude - m.magnitude) / s.magnitude,
                units=sig.units)
            sig /= sig.units
        result = signal

    # Return single object, or list of objects
    if len(result) == 1:
        return result[0]
    else:
        return result


def butter(signal, highpass_freq=None, lowpass_freq=None, order=4,
           filter_function='filtfilt', fs=1.0, axis=-1):
    """
    Butterworth filtering function for neo.AnalogSignalArray. Filter type is
    determined according to how values of `highpass_freq` and `lowpass_freq`
    are given (see Parameters section for details).

    Parameters
    ----------
    signal : AnalogSignalArray or Quantity array or NumPy ndarray
        Time series data to be filtered. When given as Quantity array or NumPy
        ndarray, the sampling frequency should be given through the keyword
        argument `fs`.
    highpass_freq, lowpass_freq : Quantity or float
        High-pass and low-pass cut-off frequencies, respectively. When given as
        float, the given value is taken as frequency in Hz.
        Filter type is determined depending on values of these arguments:
            * highpass_freq only (lowpass_freq = None):    highpass filter
            * lowpass_freq only (highpass_freq = None):    lowpass filter
            * highpass_freq < lowpass_freq:    bandpass filter
            * highpass_freq > lowpass_freq:    bandstop filter
    order : int
        Order of Butterworth filter. Default is 4.
    filter_function : string
        Filtering function to be used. Either 'filtfilt'
        (`scipy.signal.filtfilt()`) or 'lfilter' (`scipy.signal.lfilter()`). In
        most applications 'filtfilt' should be used, because it doesn't bring
        about phase shift due to filtering. Default is 'filtfilt'.
    fs : Quantity or float
        The sampling frequency of the input time series. When given as float,
        its value is taken as frequency in Hz. When the input is given as neo
        AnalogSignalArray, its attribute is used to specify the sampling
        frequency and this parameter is ignored. Default is 1.0.
    axis : int
        Axis along which filter is applied. Default is -1.

    Returns
    -------
    filtered_signal : AnalogSignalArray or Quantity array or NumPy ndarray
        Filtered input data. The shape and type is identical to those of the
        input.
    """

    def _design_butterworth_filter(Fs, hpfreq=None, lpfreq=None, order=4):
        # set parameters for filter design
        Fn = Fs / 2.
        # - filter type is determined according to the values of cut-off
        # frequencies
        if lpfreq and hpfreq:
            if hpfreq < lpfreq:
                Wn = (hpfreq / Fn, lpfreq / Fn)
                btype = 'bandpass'
            else:
                Wn = (lpfreq / Fn, hpfreq / Fn)
                btype = 'bandstop'
        elif lpfreq:
            Wn = lpfreq / Fn
            btype = 'lowpass'
        elif hpfreq:
            Wn = hpfreq / Fn
            btype = 'highpass'
        else:
            raise ValueError(
                "Either highpass_freq or lowpass_freq must be given"
            )

        # return filter coefficients
        return scipy.signal.butter(order, Wn, btype=btype)

    # design filter
    Fs = signal.sampling_rate.rescale(pq.Hz).magnitude \
        if hasattr(signal, 'sampling_rate') else fs
    Fh = highpass_freq.rescale(pq.Hz).magnitude \
        if isinstance(highpass_freq, pq.quantity.Quantity) else highpass_freq
    Fl = lowpass_freq.rescale(pq.Hz).magnitude \
        if isinstance(lowpass_freq, pq.quantity.Quantity) else lowpass_freq
    b, a = _design_butterworth_filter(Fs, Fh, Fl, order)

    # When the input is AnalogSignalArray, the axis for time index (i.e. the
    # first axis) needs to be rolled to the last
    data = np.asarray(signal)
    if isinstance(signal, neo.AnalogSignalArray):
        data = np.rollaxis(data, 0, len(data.shape))

    # apply filter
    if filter_function is 'lfilter':
        filtered_data = scipy.signal.lfilter(b, a, data, axis=axis)
    elif filter_function is 'filtfilt':
        filtered_data = scipy.signal.filtfilt(b, a, data, axis=axis)
    else:
        raise ValueError(
            "filter_func must to be either 'filtfilt' or 'lfilter'"
        )

    if isinstance(signal, neo.AnalogSignalArray):
        return signal.duplicate_with_new_array(filtered_data.T)
    elif isinstance(signal, pq.quantity.Quantity):
        return filtered_data * signal.units
    else:
        return filtered_data